class VaasAuthenticationError(BaseException):
    """Authentication Error"""


class VaasTimeoutError(BaseException):
    """Generic timeout"""


class VaasInvalidStateError(BaseException):
    """Invalid state"""


class VaasConnectionClosedError(BaseException):
    """Connection closed"""

class VaasClientError(BaseException):
    """Client Error"""
    def __init__(self, message = None) -> None:
        if message is None:
            super().__init__("Client Error")
        else:
            super().__init__(message)

class VaasServerError(BaseException):
    """Server Error"""
    def __init__(self, message = None) -> None:
        if message is None:
            super().__init__("Server Error")
        else:
            super().__init__(message)



"""Verdict-as-a-Service

:mod:`vaas` is a Python library for the VaaS-API."""
from authlib.integrations.base_client import OAuthError
from authlib.integrations.httpx_client import AsyncOAuth2Client

from .vaas_errors import VaasAuthenticationError


class ClientCredentialsGrantAuthenticator:
    """Tracing interface for Vaas"""

    def __init__(self, client_id, client_secret, token_endpoint, verify=True):
        self.client_id = client_id
        self.client_secret = client_secret
        self.token_endpoint = token_endpoint
        self.verify = verify

    async def get_token(self):
        async with AsyncOAuth2Client(
            self.client_id, self.client_secret, verify=self.verify
        ) as client:
            try:
                token = (await client.fetch_token(self.token_endpoint))["access_token"]
                return token
            except Exception as e:
                raise VaasAuthenticationError(e)



import unittest
import os
from dotenv import load_dotenv
from src.vaas import ClientCredentialsGrantAuthenticator, VaasAuthenticationError

load_dotenv()
TOKEN_URL = os.getenv("TOKEN_URL")


class ClientCredentialsGrantAuthenticatorTest(unittest.IsolatedAsyncioTestCase):
    async def test_raises_error_if_credentials_are_invalid(self):
        authenticator = ClientCredentialsGrantAuthenticator(
            "invalid_id", "invalid_secret", TOKEN_URL
        )

        with self.assertRaises(VaasAuthenticationError):
            await authenticator.get_token()

    async def test_raises_error_if_token_url_is_invalid(self):
        authenticator = ClientCredentialsGrantAuthenticator(
            "invalid_id", "invalid_secret", "isbad"
        )

        with self.assertRaises(VaasAuthenticationError):
            await authenticator.get_token()

    async def test_raises_error_if_token_url_is_wrong(self):
        authenticator = ClientCredentialsGrantAuthenticator(
            "invalid_id", "invalid_secret", "https://gdata.de"
        )

        with self.assertRaises(VaasAuthenticationError):
            await authenticator.get_token()

    async def test_raises_error_if_token_url_is_doesnotexist(self):
        authenticator = ClientCredentialsGrantAuthenticator(
            "invalid_id",
            "invalid_secret",
            "https://thishostnamedoesneverexistanywhere.de",
        )

        with self.assertRaises(VaasAuthenticationError):
            await authenticator.get_token()



"""Verdict-as-a-Service

:mod:`vaas` is a Python library for the VaaS-API."""

__version__ = "0.0.1"
__all__ = [
    "Vaas",
    "VaasTracing",
    "VaasTimeoutError",
    "VaasAuthenticationError",
    "VaasInvalidStateError",
    "VaasConnectionClosedError",
    "VaasOptions",
    "ClientCredentialsGrantAuthenticator",
    "ResourceOwnerPasswordGrantAuthenticator",
    "get_ssl_context",
]

__author__ = "G DATA CyberDefense AG <oem@gdata.de>"

from .vaas import Vaas, VaasTracing, VaasOptions, get_ssl_context
from .vaas_errors import (
    VaasTimeoutError,
    VaasAuthenticationError,
    VaasInvalidStateError,
    VaasConnectionClosedError,
)
from .client_credentials_grant_authenticator import ClientCredentialsGrantAuthenticator
from .resource_owner_password_grant_authenticator import ResourceOwnerPasswordGrantAuthenticator



import unittest
import os
from dotenv import load_dotenv
from src.vaas import ResourceOwnerPasswordGrantAuthenticator, VaasAuthenticationError, Vaas

load_dotenv()
TOKEN_URL = os.getenv("TOKEN_URL")


class ResourceOwnerPasswordGrantAuthenticatorTest(unittest.IsolatedAsyncioTestCase):

    async def test_raises_error_if_credentials_are_invalid(self):
        authenticator = ResourceOwnerPasswordGrantAuthenticator(
            "invalid_client_id",
            "invalid_user",
            "invalid_password",
            token_endpoint=TOKEN_URL)
        with self.assertRaises(VaasAuthenticationError):
            await authenticator.get_token()

    async def test_raises_error_if_token_url_is_invalid(self):
        authenticator = ResourceOwnerPasswordGrantAuthenticator(
            "invalid_client_id",
            "invalid_user",
            "invalid_password",
            token_endpoint="https://")
        with self.assertRaises(VaasAuthenticationError):
            await authenticator.get_token()

    async def test_raises_error_if_token_url_is_wrong(self):
        authenticator = ResourceOwnerPasswordGrantAuthenticator(
            "invalid_client_id",
            "invalid_user",
            "invalid_password",
            token_endpoint="gateway.production.vaas.gdatasecurity.de")
        with self.assertRaises(VaasAuthenticationError):
            await authenticator.get_token()

    async def test_raises_error_if_token_url_is_doesnotexist(self):
        authenticator = ResourceOwnerPasswordGrantAuthenticator(
            "invalid_client_id",
            "invalid_user",
            "invalid_password",
            token_endpoint="gateway.production.vaas.gdatasecurity.de/nocontenthere")
        with self.assertRaises(VaasAuthenticationError):
            await authenticator.get_token()

    async def test_for_url_with_resource_owner_password_grant_returns_malicious(self):
        token_url = os.getenv("TOKEN_URL")
        vaas_url = os.getenv("VAAS_URL")
        client_id = os.getenv("VAAS_CLIENT_ID")
        username = os.getenv("VAAS_USER_NAME")
        password = os.getenv("VAAS_PASSWORD")

        if token_url is None:
            token_url = "https://account.gdata.de/realms/vaas-production/protocol/openid-connect/token"
        if vaas_url is None:
            vaas_url = "wss://gateway.production.vaas.gdatasecurity.de"

        authenticator = ResourceOwnerPasswordGrantAuthenticator(
            client_id=client_id,
            user_name=username,
            password=password,
            token_endpoint=token_url
        )
        async with Vaas(url=vaas_url) as vaas:
            await vaas.connect(await authenticator.get_token())
            url = "https://secure.eicar.org/eicar.com"
            verdict = await vaas.for_url(url)
            print(f"Url {url} is detected as {verdict.get('Verdict')}")
        self.assertEqual(verdict.get("Verdict"), "Malicious")



"""Verdict-as-a-Service

:mod:`vaas` is a Python library for the VaaS-API."""
import hashlib
import json
import time
import uuid
from typing import Optional, TypedDict, Literal
import asyncio
from asyncio import Future
import ssl
from urllib.parse import urlparse
import aiofiles
from jwt import PyJWT
import httpx
import websockets.client
from .vaas_errors import (
    VaasInvalidStateError,
    VaasConnectionClosedError,
    VaasTimeoutError, VaasClientError, VaasServerError,
)

TIMEOUT = 60
HTTP2 = False
# TODO: Set to default of 5 once Vaas upload endpoint is 100% streaming
UPLOAD_TIMEOUT = 600


class VaasTracing:
    """Tracing interface for Vaas"""

    def trace_hash_request(self, elapsed_in_seconds):
        """Trace hash request in seconds"""

    def trace_upload_request(self, elapsed_in_seconds, file_size):
        """Trace upload request in seconds"""

    def trace_url_request(self, elapsed_in_seconds):
        """Trace url request in seconds"""

    def trace_hash_request_timeout(self):
        """Trace timeout while waiting for hash verdict"""

    def trace_upload_result_timeout(self, file_size):
        """Trace timeout while waiting for verdict for uploaded file"""

    def trace_upload_timeout(self, file_size):
        """Trace upload timeout"""


class VaasOptions:
    """Configure behaviour of VaaS"""

    def __init__(self):
        self.use_cache = True
        self.use_hash_lookup = True

class VaasVerdict(TypedDict):
    Sha256: str
    "The SHA256 hash of the file"

    Guid: str

    Verdict: Literal["Clean", "Malicious", "Unknown", "Pup"]
    
    Detection: Optional[str]
    "Name of the detected malware if found"
    
    FileType: Optional[str]
    "The file type of the file"

    MimeType: Optional[str]
    "The mime type of the file"

def hash_file(filename):
    """Return sha256 hash for file"""
    block_size = 65536

    h_sha256 = hashlib.sha256()

    with open(filename, "rb") as file:
        buffer = file.read(block_size)
        while len(buffer) > 0:
            h_sha256.update(buffer)
            buffer = file.read(block_size)

    return h_sha256.hexdigest()


def is_ssl_url(url):
    """check if url is wss"""
    parsed_url = urlparse(url)
    return parsed_url.scheme == "wss"


def get_ssl_context(url, verify):
    """return ssl context for websockets"""
    if not is_ssl_url(url):
        return None
    if verify:
        return ssl.create_default_context()
    return ssl._create_unverified_context()  # pylint: disable=W0212


def problem_details_to_error(problem_details):
    type = problem_details.get("type")
    details = problem_details.get("details")
    if type == "VaasClientException":
        return VaasClientError(details)
    return VaasServerError(details)


def map_response(verdict_response) -> VaasVerdict:
    return {
        "Sha256": verdict_response.get("sha256"),
        "Guid": verdict_response.get("guid"),
        "Verdict": verdict_response.get("verdict"),
        "Detection": verdict_response.get("detection"),
        "FileType": verdict_response.get("file_type"),
        "MimeType": verdict_response.get("mime_type")
    }


class Vaas:
    """Verdict-as-a-Service client"""

    def __init__(
        self,
        tracing=VaasTracing(),
        options=VaasOptions(),
        url="wss://gateway.production.vaas.gdatasecurity.de",
    ):
        self.tracing = tracing
        self.loop_result = None
        self.websocket = None
        self.session_id = None
        self.results = {}
        self.httpx_client: Optional[httpx.AsyncClient] = None
        self.options = options
        self.url = url

    def get_authenticated_websocket(self):
        """Get authenticated websocket"""
        if self.websocket is None:
            raise VaasInvalidStateError("connect() was not called")
        if not self.websocket.open:
            raise VaasConnectionClosedError(
                "connection closed or connect() was not awaited"
            )
        if self.session_id is None:
            raise VaasConnectionClosedError("connect() was not awaited")
        return self.websocket

    async def connect(self, token, verify=True, websocket=None):
        """Connect to VaaS

        token -- OpenID Connect token signed by a trusted identity provider
        """
        ssl_context = get_ssl_context(self.url, verify)
        if websocket is not None:
            self.websocket = websocket
        else:
            self.websocket = await websockets.client.connect(self.url, ssl=ssl_context)

        authenticate_request = {"kind": "AuthRequest", "token": token}
        await self.websocket.send(json.dumps(authenticate_request))

        authentication_response = json.loads(await self.websocket.recv())
        if not authentication_response.get("success", False):
            raise Exception("Authentication failed")
        self.session_id = authentication_response["session_id"]

        self.loop_result = asyncio.ensure_future(
            self.__receive_loop()
        )  # fire and forget async_foo()

        self.httpx_client = httpx.AsyncClient(http2=HTTP2, verify=verify)

    async def close(self):
        """Close the connection"""
        if self.websocket is not None:
            await self.websocket.close()
        if self.loop_result is not None:
            await self.loop_result
        if self.httpx_client is not None:
            await self.httpx_client.aclose()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, traceback):
        await self.close()

    async def for_sha256(self, sha256, verdict_request_attributes=None, guid=None) -> VaasVerdict:
        """Returns the verdict for a SHA256 checksum"""
        verdict_response = await self.__for_sha256(
            sha256, verdict_request_attributes, guid
        )
        return map_response(verdict_response)

    async def __for_stream(self, verdict_request_attributes=None, guid=None):
        if verdict_request_attributes is not None and not isinstance(
            verdict_request_attributes, dict
        ):
            raise TypeError("verdict_request_attributes has to be dict(str, str)")

        websocket = self.get_authenticated_websocket()
        start = time.time()
        guid = guid or str(uuid.uuid4())
        verdict_request = {
            "kind": "VerdictRequestForStream",
            "session_id": self.session_id,
            "guid": guid,
            "use_hash_lookup": self.options.use_hash_lookup,
            "use_cache": self.options.use_cache,
            "verdict_request_attributes": verdict_request_attributes,
        }
        response_message = self.__response_message_for_guid(guid)
        await websocket.send(json.dumps(verdict_request))

        try:
            result = await asyncio.wait_for(response_message, timeout=TIMEOUT)
        except asyncio.TimeoutError as ex:
            self.tracing.trace_hash_request_timeout()
            raise VaasTimeoutError() from ex

        self.tracing.trace_hash_request(time.time() - start)
        return result

    async def __for_sha256(self, sha256, verdict_request_attributes=None, guid=None):
        if verdict_request_attributes is not None and not isinstance(
            verdict_request_attributes, dict
        ):
            raise TypeError("verdict_request_attributes has to be dict(str, str)")

        websocket = self.get_authenticated_websocket()
        start = time.time()
        guid = guid or str(uuid.uuid4())
        verdict_request = {
            "kind": "VerdictRequest",
            "sha256": sha256,
            "session_id": self.session_id,
            "guid": guid,
            "use_hash_lookup": self.options.use_hash_lookup,
            "use_cache": self.options.use_cache,
            "verdict_request_attributes": verdict_request_attributes,
        }
        response_message = self.__response_message_for_guid(guid)
        await websocket.send(json.dumps(verdict_request))

        try:
            result = await asyncio.wait_for(response_message, timeout=TIMEOUT)
        except asyncio.TimeoutError as ex:
            self.tracing.trace_hash_request_timeout()
            raise VaasTimeoutError() from ex

        self.tracing.trace_hash_request(time.time() - start)
        return result

    def __response_message_for_guid(self, guid):
        result = Future()
        self.results[guid] = result
        return result

    async def __receive_loop(self):
        try:
            websocket = self.get_authenticated_websocket()
        except VaasConnectionClosedError as e:
            # Fix for Python >= 3.12: Connection has already been closed again. End loop.
            return
        try:
            async for message in websocket:
                vaas_message = json.loads(message)
                message_kind = vaas_message.get("kind")
                guid = vaas_message.get("guid")
                if message_kind == "VerdictResponse":
                    future = self.results.get(guid)
                    if future is not None:
                        future.set_result(vaas_message)
                if message_kind == "Error":
                    problem_details = vaas_message.get("problem_details")
                    if guid is None or problem_details is None:
                        # Error: Server sent guid we are not waiting for, or problem details are null, ignore it
                        continue
                    future = self.results.get(guid)
                    if future is not None:
                        future.set_exception(problem_details_to_error(problem_details))
        except Exception as error:
            raise VaasConnectionClosedError(error) from error

    async def for_buffer(self, buffer, verdict_request_attributes=None, guid=None) -> VaasVerdict:
        """Returns the verdict for a buffer"""

        loop = asyncio.get_running_loop()
        sha256 = await loop.run_in_executor(
            None, lambda: hashlib.sha256(buffer).hexdigest()
        )

        verdict_response = await self.__for_sha256(
            sha256, verdict_request_attributes, guid
        )
        verdict = verdict_response.get("verdict")

        if verdict == "Unknown":
            verdict_response = await self._for_unknown_buffer(
                verdict_response, buffer, len(buffer)
            )

        return map_response(verdict_response)

    async def _for_unknown_buffer(self, response, buffer, buffer_len):
        start = time.time()
        guid = response.get("guid")
        token = response.get("upload_token")
        url = response.get("url")
        response_message = self.__response_message_for_guid(guid)
        await self.__upload(token, url, buffer, buffer_len)
        try:
            verdict_response = await asyncio.wait_for(response_message, timeout=TIMEOUT)
        except asyncio.TimeoutError as ex:
            self.tracing.trace_upload_result_timeout(buffer_len)
            raise VaasTimeoutError() from ex
        self.tracing.trace_upload_request(time.time() - start, buffer_len)
        return verdict_response

    async def for_stream(self, asyncBufferedReader, len, verdict_request_attributes=None, guid=None) -> VaasVerdict:
        """Returns the verdict for a file"""

        verdict_response = await self.__for_stream(
            verdict_request_attributes, guid
        )
        guid = verdict_response.get("guid")
        token = verdict_response.get("upload_token")
        url = verdict_response.get("url")
        verdict = verdict_response.get("verdict")

        if verdict != "Unknown":
            raise VaasServerError("server returned verdict without receiving content")

        if token == None:
            raise VaasServerError("VerdictResponse missing UploadToken for stream upload")

        if url == None:
            raise VaasServerError("VerdictResponse missing URL for stream upload")

        start = time.time()
        response_message = self.__response_message_for_guid(guid)
        await self.__upload(token, url, asyncBufferedReader, len)
        try:
            verdict_response = await asyncio.wait_for(response_message, timeout=TIMEOUT)
        except asyncio.TimeoutError as ex:
            self.tracing.trace_upload_result_timeout(len)
            raise VaasTimeoutError() from ex
        self.tracing.trace_upload_request(time.time() - start, len)

        return map_response(verdict_response)

    async def for_file(self, path, verdict_request_attributes=None, guid=None) -> VaasVerdict:
        """Returns the verdict for a file"""

        loop = asyncio.get_running_loop()
        sha256 = await loop.run_in_executor(None, lambda: hash_file(path))

        verdict_response = await self.__for_sha256(
            sha256, verdict_request_attributes, guid
        )
        verdict = verdict_response.get("verdict")

        if verdict == "Unknown":
            async with aiofiles.open(path, mode="rb") as file:
                buffer = await file.read()
                verdict_response = await self._for_unknown_buffer(
                    verdict_response, buffer, len(buffer)
                )

        return map_response(verdict_response)

    async def __upload(self, token, upload_uri, buffer_or_file, content_length):
        jwt = PyJWT()
        decoded_token = jwt.decode(token, options={"verify_signature": False})
        try:
            await self.httpx_client.put(
                url=upload_uri,
                content=buffer_or_file,
                headers={
                    "Authorization": token,
                    "Content-Length": str(content_length),
                },
                timeout=UPLOAD_TIMEOUT,
            )
        except httpx.TimeoutException as ex:
            self.tracing.trace_upload_timeout(content_length)
            raise VaasTimeoutError() from ex

    async def for_url(self, url, verdict_request_attributes=None, guid=None) -> VaasVerdict:
        """Returns the verdict for a file from an url"""
        if verdict_request_attributes is not None and not isinstance(
            verdict_request_attributes, dict
        ):
            raise TypeError("verdict_request_attributes has to be dict(str, str)")

        websocket = self.get_authenticated_websocket()
        start = time.time()
        guid = guid or str(uuid.uuid4())
        verdict_request_for_url = {
            "kind": "VerdictRequestForUrl",
            "url": url,
            "session_id": self.session_id,
            "guid": guid,
            "use_hash_lookup": self.options.use_hash_lookup,
            "use_cache": self.options.use_cache,
            "verdict_request_attributes": verdict_request_attributes,
        }
        response_message = self.__response_message_for_guid(guid)
        await websocket.send(json.dumps(verdict_request_for_url))

        try:
            result = await asyncio.wait_for(response_message, timeout=TIMEOUT)
        except asyncio.TimeoutError as ex:
            self.tracing.trace_hash_request_timeout()
            raise VaasTimeoutError() from ex

        self.tracing.trace_url_request(time.time() - start)

        return map_response(result)



# pylint: disable=C0114,C0116,C0115
import base64
import hashlib
import os
import unittest
from unittest.mock import MagicMock, ANY
import uuid

import websockets.client
from dotenv import load_dotenv

from src.vaas import Vaas, VaasTracing, VaasOptions, ClientCredentialsGrantAuthenticator
from src.vaas import get_ssl_context
from src.vaas.vaas import hash_file
from src.vaas.vaas_errors import VaasConnectionClosedError, VaasInvalidStateError, VaasClientError
import httpx

load_dotenv()

CLIENT_ID = os.getenv("CLIENT_ID")
CLIENT_SECRET = os.getenv("CLIENT_SECRET")
TOKEN_URL = os.getenv("TOKEN_URL")
VAAS_URL = os.getenv("VAAS_URL")
SSL_VERIFICATION = os.getenv("SSL_VERIFICATION", "True").lower() in ["true", "1"]

EICAR_BASE64 = "WDVPIVAlQEFQWzRcUFpYNTQoUF4pN0NDKTd9JEVJQ0FSLVNUQU5EQVJELUFOVElWSVJVUy1URVNULUZJTEUhJEgrSCo="


async def create_and_connect(tracing=VaasTracing(), options=VaasOptions()):
    authenticator = ClientCredentialsGrantAuthenticator(
        CLIENT_ID, CLIENT_SECRET, TOKEN_URL, SSL_VERIFICATION
    )
    vaas = Vaas(tracing=tracing, options=options, url=VAAS_URL)

    token = await authenticator.get_token()
    await vaas.connect(token, verify=SSL_VERIFICATION)
    return vaas


def get_disabled_options():
    options = VaasOptions()
    options.use_cache = False
    options.use_hash_lookup = False
    return options


class VaasTest(unittest.IsolatedAsyncioTestCase):
    async def test_raises_error_if_token_is_invalid(self):
        async with Vaas() as vaas:
            token = "ThisIsAnInvalidToken"
            with self.assertRaises(Exception):
                await vaas.connect(token)

    async def test_connects(self):
        async with await create_and_connect():
            pass

    async def test_for_sha256_returns_clean_for_clean_sha256(self):
        async with await create_and_connect() as vaas:
            verdict = await vaas.for_sha256(
                "cd617c5c1b1ff1c94a52ab8cf07192654f271a3f8bad49490288131ccb9efc1e"
            )
            self.assertEqual(verdict["Verdict"], "Clean")
            self.assertEqual(
                verdict["Sha256"].casefold(),
                "cd617c5c1b1ff1c94a52ab8cf07192654f271a3f8bad49490288131ccb9efc1e".casefold(),
            )

    async def test_use_for_sha256_when_connection_already_closed(self):
        authenticator = ClientCredentialsGrantAuthenticator(
            CLIENT_ID, CLIENT_SECRET, TOKEN_URL, SSL_VERIFICATION
        )
        vaas = Vaas(tracing=VaasTracing(), options=VaasOptions(), url=VAAS_URL)
        ssl_context = get_ssl_context(VAAS_URL, SSL_VERIFICATION)
        websocket = await websockets.client.connect(VAAS_URL, ssl=ssl_context)
        await vaas.connect(
            await authenticator.get_token(),
            websocket=websocket,
            verify=SSL_VERIFICATION,
        )
        await websocket.close()
        with self.assertRaises(VaasConnectionClosedError):
            await vaas.for_sha256(
                "275a021bbfb6489e54d471899f7db9d1663fc695ec2fe2a2c4538aabf651fd0f"
            )

    async def test_use_for_sha256_if_not_connected(self):
        vaas = Vaas(tracing=VaasTracing(), options=VaasOptions(), url=VAAS_URL)
        with self.assertRaises(VaasInvalidStateError):
            await vaas.for_sha256(
                "275a021bbfb6489e54d471899f7db9d1663fc695ec2fe2a2c4538aabf651fd0f"
            )

    async def test_for_stream_eicar_form_url_returns_malicious(self):
        async with await create_and_connect() as vaas:
            guid = str(uuid.uuid4())
            async with httpx.AsyncClient() as client:
                response = await client.get("https://secure.eicar.org/eicar.com")
                content_length = response.headers["Content-Length"]
                verdict = await vaas.for_stream(
                    response.aiter_bytes(),
                    content_length,
                    guid=guid
                )
                self.assertEqual(verdict["Verdict"], "Malicious")
                self.assertEqual(verdict["Guid"].casefold(), guid)

    async def test_for_sha256_returns_malicious_for_eicar(self):
        async with await create_and_connect() as vaas:
            guid = str(uuid.uuid4())
            verdict = await vaas.for_sha256(
                "275a021bbfb6489e54d471899f7db9d1663fc695ec2fe2a2c4538aabf651fd0f",
                guid=guid
            )
            self.assertEqual(verdict["Verdict"], "Malicious")
            self.assertEqual(
                verdict["Sha256"].casefold(),
                "275a021bbfb6489e54d471899f7db9d1663fc695ec2fe2a2c4538aabf651fd0f".casefold(),
            self.assertEqual(verdict["Guid"].casefold(), guid)
            )

    async def test_for_sha256_returns_pup_for_amtso(self):
        async with await create_and_connect() as vaas:
            guid = str(uuid.uuid4())
            verdict = await vaas.for_sha256(
                "d6f6c6b9fde37694e12b12009ad11ab9ec8dd0f193e7319c523933bdad8a50ad",
                guid=guid
            )
            self.assertEqual(verdict["Verdict"], "Pup")
            self.assertEqual(
                verdict["Sha256"].casefold(),
                "d6f6c6b9fde37694e12b12009ad11ab9ec8dd0f193e7319c523933bdad8a50ad".casefold(),
            )
            self.assertEqual(verdict["Guid"].casefold(), guid)

    async def test_for_buffer_returns_malicious_for_eicar(self):
        async with await create_and_connect() as vaas:
            buffer = base64.b64decode(EICAR_BASE64)
            sha256 = hashlib.sha256(buffer).hexdigest()
            guid = str(uuid.uuid4())
            verdict = await vaas.for_buffer(buffer, guid=guid)
            self.assertEqual(verdict["Verdict"], "Malicious")
            self.assertEqual(verdict["Sha256"].casefold(), sha256.casefold())
            self.assertEqual(verdict["Guid"].casefold(), guid)

    async def test_for_stream_returns_malicious_for_eicar(self):
        async with await create_and_connect() as vaas:
            buffer = base64.b64decode(EICAR_BASE64)
            guid = str(uuid.uuid4())
            verdict = await vaas.for_buffer(buffer, guid=guid)
            self.assertEqual(verdict["Verdict"], "Malicious")
            self.assertEqual(verdict["Guid"].casefold(), guid)

    async def test_for_buffer_returns_unknown_for_random_buffer(self):
        async with await create_and_connect() as vaas:
            buffer = os.urandom(1024)
            sha256 = hashlib.sha256(buffer).hexdigest()
            guid = str(uuid.uuid4())
            verdict = await vaas.for_buffer(buffer, guid=guid)
            self.assertEqual(verdict["Verdict"], "Clean")
            self.assertEqual(verdict["Sha256"].casefold(), sha256.casefold())
            self.assertEqual(verdict["Guid"].casefold(), guid)

    async def test_for_file_returns_verdict(self):
        async with await create_and_connect() as vaas:
            with open("eicar.txt", "wb") as f:
                f.write(base64.b64decode(EICAR_BASE64))
            sha256 = hash_file("eicar.txt")
            guid = str(uuid.uuid4())
            verdict = await vaas.for_file("eicar.txt", guid=guid)
            self.assertEqual(verdict["Verdict"], "Malicious")
            self.assertEqual(verdict["Sha256"].casefold(), sha256.casefold())
            self.assertEqual(verdict["Guid"].casefold(), guid)

    async def test_for_file_returns_verdict_if_no_cache_or_shed(self):
        options = get_disabled_options()

        async with await create_and_connect(options=options) as vaas:
            with open("eicar.txt", "wb") as f:
                f.write(base64.b64decode(EICAR_BASE64))
            sha256 = hash_file("eicar.txt")
            guid = str(uuid.uuid4())
            verdict = await vaas.for_file("eicar.txt", guid=guid)
            self.assertEqual(verdict["Verdict"], "Malicious")
            self.assertEqual(verdict["Sha256"].casefold(), sha256.casefold())
            self.assertEqual(verdict["Guid"].casefold(), guid)

    async def test_for_url_returns_malicious_for_eicar(self):
        options = get_disabled_options()
        async with await create_and_connect(options=options) as vaas:
            guid = str(uuid.uuid4())
            verdict = await vaas.for_url("https://secure.eicar.org/eicarcom2.zip", guid=guid)
            self.assertEqual(verdict["Verdict"], "Malicious")
            self.assertEqual(verdict["Guid"].casefold(), guid)

    async def test_for_url_without_shed_and_cache_returns_clean_for_robots_txt(self):
        options = get_disabled_options()
        async with await create_and_connect(options=options) as vaas:
            guid = str(uuid.uuid4())
            verdict = await vaas.for_url("https://www.gdata.de/robots.txt", guid=guid)
            self.assertEqual(verdict["Verdict"], "Clean")
            self.assertEqual(verdict["Guid"].casefold(), guid)

    async def test_for_url_without_cache_returns_clean_for_robots_txt(self):
        options = VaasOptions()
        options.use_cache = False
        options.use_hash_lookup = True
        async with await create_and_connect(options=options) as vaas:
            guid = str(uuid.uuid4())
            verdict = await vaas.for_url("https://www.gdata.de/robots.txt", guid=guid)
            self.assertEqual(verdict["Verdict"], "Clean")
            self.assertEqual(verdict["Guid"].casefold(), guid)

    async def test_for_url__with_url_with_status_code_4xx__raises_VaasClientError(self):
        options = get_disabled_options()
        async with await create_and_connect(options=options) as vaas:
            with self.assertRaises(VaasClientError, msg="Call failed with status code 404 (Not Found): GET https://gateway.production.vaas.gdatasecurity.de/swagger/nocontenthere") as error:
                await vaas.for_url("https://gateway.production.vaas.gdatasecurity.de/swagger/nocontenthere")
            self.assertEqual(str(error.msg), "Call failed with status code 404 (Not Found): GET https://gateway.production.vaas.gdatasecurity.de/swagger/nocontenthere")

    async def test_for_buffer_traces(self):
        tracing = VaasTracing()
        tracing.trace_hash_request = MagicMock()
        tracing.trace_upload_request = MagicMock()
        async with await create_and_connect(tracing=tracing) as vaas:
            buffer = os.urandom(1024)
            sha256 = hashlib.sha256(buffer).hexdigest()
            guid = str(uuid.uuid4())
            verdict = await vaas.for_buffer(buffer, guid=guid)
            self.assertEqual(verdict["Verdict"], "Clean")
            self.assertEqual(verdict["Sha256"].casefold(), sha256.casefold())
            self.assertEqual(verdict["Guid"].casefold(), guid)
            tracing.trace_hash_request.assert_called_with(ANY)
            tracing.trace_upload_request.assert_called_with(ANY, 1024)

    async def test_for_empty_buffer_returns_clean(self):
        async with await create_and_connect() as vaas:
            buffer = bytes("", "utf-8")
            sha256 = hashlib.sha256(buffer).hexdigest()
            guid = str(uuid.uuid4())
            verdict = await vaas.for_buffer(buffer, guid=guid)
            self.assertEqual(verdict["Verdict"], "Clean")
            self.assertEqual(verdict["Sha256"].casefold(), sha256.casefold())
            self.assertEqual(verdict["Guid"].casefold(), guid)

    async def test_for_url_returns_detections_and_mime_type(self):
        options = get_disabled_options()
        async with await create_and_connect(options=options) as vaas:
            guid = str(uuid.uuid4())
            verdict = await vaas.for_url("https://secure.eicar.org/eicar.com.txt", guid=guid)
            self.assertEqual(verdict["Verdict"], "Malicious")
            self.assertIsNotNone(verdict["Detection"])
            self.assertEqual(verdict['FileType'], "EICAR virus test files")
            self.assertEqual(verdict['MimeType'], "text/plain")


if __name__ == "__main__":
    unittest.main()
